package mi.hdm.shoppingList;

import mi.hdm.exceptions.InvalidIngredientException;
import mi.hdm.recipes.Ingredient;
import mi.hdm.recipes.Recipe;
import mi.hdm.recipes.RecipeManager;
import mi.hdm.recipes.ValidObjectsPool;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.Map;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

class ShoppingListTest {
    private final static RecipeManager recipeManager = ValidObjectsPool.getRecipeManager();
    private final static ShoppingList underTest = new ShoppingList(recipeManager);
    private final static Ingredient ing1 = ValidObjectsPool.getValidIngredientOne();
    private final static Ingredient ing2 = ValidObjectsPool.getValidIngredientTwo();

    private final static Recipe r1 = ValidObjectsPool.getValidRecipeOne();

    @BeforeEach
    public void setup() {
        recipeManager.clear();
        underTest.clear();
    }

    @Test
    public void shouldAddIngredient() {
        //given
        Map<String, Boolean> expected = Map.of(ing1.getUniqueCode(), false);

        //when
        underTest.addToShoppingList(ing1);

        //expect
        assertEquals(expected, underTest.getList());
    }

    @Test
    public void shouldAddAllIngredients() {
        //given
        Map<String, Boolean> expected = Map.of(ing1.getUniqueCode(), false, ing2.getUniqueCode(), false);

        //when
        underTest.addAllToShoppingList(r1);

        //expect
        assertEquals(expected, underTest.getList());
    }

    @Test
    public void shouldAllAllIngredientsRecursively() {
        //given
        Recipe r1 = new Recipe("parent", Map.of(ValidObjectsPool.getValidIngredientOne(), 100), "Desc", List.of("prep"), List.of(), 10);
        Recipe r2 = new Recipe("child", Map.of(r1, 100), "Desc", List.of("prep"), List.of(), 10);
        Map<String, Boolean> expected = Map.of(ValidObjectsPool.getValidIngredientOne().getUniqueCode(), false);
        recipeManager.addRecipe(r1);
        recipeManager.addRecipe(r2);

        //when
        underTest.addAllToShoppingList(r2);

        //expect
        assertEquals(expected, underTest.getList());
    }

    @Test
    public void testSetStatusByObject() {
        //given
        Map<String, Boolean> expected = Map.of(ing1.getUniqueCode(), true);
        underTest.addToShoppingList(ing1);

        //when
        underTest.flipStatus(ing1);

        //expect
        assertEquals(expected, underTest.getList());
        assertThrows(InvalidIngredientException.class, () -> underTest.flipStatus(ing2));
        assertThrows(NullPointerException.class, () -> underTest.flipStatus((Ingredient) null));
    }

    @Test
    public void testSetStatusByKey() {
        //given
        Map<String, Boolean> expected = Map.of(ing1.getUniqueCode(), true);
        underTest.addToShoppingList(ing1);

        //when
        underTest.flipStatus(ing1.getUniqueCode());

        //expect
        assertEquals(expected, underTest.getList());
        assertThrows(InvalidIngredientException.class, () -> underTest.flipStatus(ing2));
        assertThrows(NullPointerException.class, () -> underTest.flipStatus((Ingredient) null));
    }

    @Test
    public void shouldRemoveAllBoughtItems() {
        //given
        Map<String, Boolean> expected = Map.of();
        underTest.addToShoppingList(ing1);
        underTest.flipStatus(ing1);

        //when
        underTest.removeAllBought();

        //expect
        assertEquals(expected, underTest.getList());
    }
}